use libc::{
    msgctl, msgget, msqid_ds, IPC_CREAT, IPC_INFO, IPC_NOWAIT, IPC_RMID, MSG_INFO,g MSG_STAT,
};
use nix::errno::{errno, Errno};
use std::{marker::PhantomData, mem::MaybeUninit, ptr};
use thiserror::Error;

#[derive(Debug, Error)]
pub enum SysvMqError {
    #[error("SysV message queue: {0}")]
    ErrnoError(&'static str),
}

#[allow(clippy::doc_markdown)]
/// Unlink (delete) an existing SysV IPC message queue.
///
/// # Errors
///
/// Return an `SysvMqError` when no queue with the given key can be
/// found.
pub fn unlink_id(id: i32) -> Result<(), SysvMqError> {
    let res = unsafe { msgctl(id, IPC_RMID, ptr::null::<msqid_ds>().cast_mut()) };

    match res {
        -1 => Err(SysvMqError::ErrnoError(Errno::from_i32(errno()).desc())),
        _ => Ok(()),
    }
}

#[allow(clippy::doc_markdown)]
/// Get the id of an existing SysV IPC message queue by passing its
/// key.
///
/// # Errors
///
/// Return an `SysvMqError` when no queue with the given key can be
/// found.
pub fn id_from_key(key: i32) -> Result<i32, SysvMqError> {
    let id = unsafe { msgget(key, 0) };

    match id {
        -1 => Err(SysvMqError::ErrnoError(Errno::from_i32(errno()).desc())),
        id => Ok(id),
    }
}

#[allow(clippy::doc_markdown)]
/// Get the info about an existing SysV IPC message queue.
pub fn ipc_info(id: i32) {
    let mut ipc_info = MaybeUninit::<msqid_ds>::uninit();

    let ipc_info = unsafe {
        msgctl(id, IPC_INFO, ipc_info.as_mut_ptr());
        ipc_info.assume_init()
    };

    println!("info: {ipc_info:?}");
}

#[allow(clippy::doc_markdown)]
/// Get the stats about an existing SysV IPC message queue.
pub fn stat_info(id: i32) {
    let mut stat_info = MaybeUninit::<msqid_ds>::uninit();

    let stat_info = unsafe {
        msgctl(id, MSG_STAT, stat_info.as_mut_ptr());
        stat_info.assume_init()
    };

    println!("info: {stat_info:?}");
}

#[allow(clippy::doc_markdown)]
/// Get the message info about an existing SysV IPC message queue.
pub fn msg_info(id: i32) {
    let mut msg_info = MaybeUninit::<msqid_ds>::uninit();

    let msg_info = unsafe {
        msgctl(id, MSG_INFO, msg_info.as_mut_ptr());
        msg_info.assume_init()
    };

    println!("info: {msg_info:?}");
}

pub struct SysvMq<T> {
    pub id: i32,
    pub key: i32,
    message_mask: i32,
    mode: i32,
    types: PhantomData<T>,
}

impl<T> SysvMq<T> {
    /// Create a new message queye with the given key.
    ///
    /// # Errors
    ///
    /// Return an `SysvMqError` when no queue with the given key can be
    /// created.
    pub fn create(&mut self, key: i32) -> Result<&Self, SysvMqError> {
        self.key = key;
        self.id = unsafe { msgget(self.key, IPC_CREAT | self.mode) };

        match self.id {
            -1 => Err(SysvMqError::ErrnoError(Errno::from_i32(errno()).desc())),
            _ => Ok(self),
        }
    }

    /// Open an existing message queye with the given key.
    ///
    /// # Errors
    ///
    /// Return an `SysvMqError` when no queue with the given key can be
    /// found.
    pub fn open(mut self, key: i32) -> Result<Self, SysvMqError> {
        self.key = key;
        self.id = unsafe { msgget(self.key, self.mode) };

        match self.id {
            -1 => Err(SysvMqError::ErrnoError(Errno::from_i32(errno()).desc())),
            _ => Ok(self),
        }
    }

    pub fn mode(&mut self, mode: i32) -> &Self {
        self.mode = mode;
        self
    }

    pub fn nonblocking(&mut self) -> &Self {
        self.message_mask |= IPC_NOWAIT;
        self
    }

    #[must_use]
    pub const fn new() -> Self {
        Self {
            id: -1,
            key: 0,
            message_mask: 0,
            mode: 0o644,
            types: PhantomData,
        }
    }
}

impl<T> Default for SysvMq<T> {
    fn default() -> Self {
        Self::new()
    }
}
